"""
Handles Connection duties (reconnecting etc.) transparently.
"""
import logging
import struct

from bluepy import btle

REQUIREMENTS = ['bluepy>=1.0.4']

DEFAULT_TIMEOUT = 1

_LOGGER = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class BTLEConnection(btle.DefaultDelegate):
    """Representation of a BTLE Connection."""

    def __init__(self, mac):
        """Initialize the connection."""
        btle.DefaultDelegate.__init__(self)

        self._mac = mac
        self._conn = btle.Peripheral()

        self._callback = {}

    def __del__(self):
        """Destructor - make sure the connection is disconnected."""
        self.disconnect()

    def handleNotification(self, handle, data):
        """Handle Callback from a Bluetooth (GATT) request."""
        if (handle in self._callback):
            self._callback[handle](data)

    def connect(self, error=False):
        """Connect to the Bluetooth thermostat."""
        _LOGGER.info("BTLEConnection: connecting to " + self._mac)
        try:
            self._conn.connect(self._mac)
            self._conn.withDelegate(self)
        except btle.BTLEException:
            if (error):
                raise
            else:
                self.disconnect()
                self.connect(True)

    def disconnect(self):
        """Close the Bluetooth connection."""
        self._conn.disconnect()

    @property
    def mac(self):
        """Return the MAC address of the connected device."""
        return self._mac

    def set_callback(self, handle, function):
        """Set the callback for a Notification handle. It will be called with the parameter data, which is binary."""
        self._callback[handle] = function

    def write_request(self, handle, value, timeout=DEFAULT_TIMEOUT):
        """Write a GATT Command with callback."""
        self.write_command(handle, value, timeout, True)

    def write_request_raw(self, handle, value, timeout=DEFAULT_TIMEOUT):
        """Write a GATT Command with callback - no utf-8."""
        self.write_command_raw(handle, value, timeout, True)

    def write_command(self, handle, value, timeout=DEFAULT_TIMEOUT, wait_for_it=False):
        """Write a GATT Command without callback."""
        self.write_command_raw(handle, value.encode('utf-8'), timeout, wait_for_it)

    def write_command_raw(self, handle, value, timeout=DEFAULT_TIMEOUT, wait_for_it=False, exception=False):
        """Write a GATT Command without callback - not utf-8."""
        try:
            self._conn.writeCharacteristic(handle, value, wait_for_it)
            if wait_for_it:
                while self._conn.waitForNotifications(timeout):
                    continue
        except btle.BTLEException:
            if exception is False:
                self.disconnect()
                self.connect()
                self.write_command_raw(handle, value, wait_for_it, True)

    @staticmethod
    def pack_byte(byte):
        """Pack a byte."""
        return struct.pack('B', byte)



"""
Support for eq3 Bluetooth Smart thermostats.

All temperatures in Celsius.

To get the current state, update() has to be called for powersaving reasons.
"""

import struct
from datetime import datetime
from ..lib.connection import BTLEConnection

PROP_WRITE_HANDLE = 0x411
PROP_NTFY_HANDLE = 0x421

PROP_ID_QUERY = 0
PROP_ID_RETURN = 1
PROP_INFO_QUERY = 3
PROP_INFO_RETURN = 2
PROP_SCHEDULE_QUERY = 0x20
PROP_SCHEDULE_RETURN = 0x21

PROP_TEMPERATURE_WRITE = 0x41


# pylint: disable=too-many-instance-attributes
class EQ3BTSmartThermostat:
    """Representation of a EQ3 Bluetooth Smart thermostat."""

    def __init__(self, _mac):
        """Initialize the thermostat."""

        self._target_temperature = -1
        self._mode = -1

        self._conn = BTLEConnection(_mac)

        self._conn.set_callback(PROP_NTFY_HANDLE, self.handle_notification)
        self._conn.connect()

        self._set_time()
        self.update()

    def __str__(self):
        return "MAC: " + self._conn.mac + " Mode: " + str(self.mode) + " = " + self.mode_readable + " T: " + str(self.target_temperature)

    def handle_notification(self, data):
        """Handle Callback from a Bluetooth (GATT) request."""
        if data[0] == PROP_INFO_RETURN:
            self._mode = data[2] & 1
            self._target_temperature = data[5] / 2.0

# TO BE TESTED
    def _set_time(self):
        """Set the correct time into the thermostat."""
        time = datetime.now()
        value = struct.pack('BBBBBB', int(time.strftime("%y")),
                            time.month, time.day, time.hour,
                            time.minute, time.second)
        self._conn.write_command_raw(PROP_WRITE_HANDLE, value)

    @property
    def target_temperature(self):
        """Return the temperature we try to reach."""
        return self._target_temperature

    @target_temperature.setter
    def target_temperature(self, temperature):
        """Set new target temperature."""
        value = struct.pack('BB', PROP_TEMPERATURE_WRITE, int(temperature * 2))
        # Returns INFO_QUERY, so use that
        self._conn.write_request_raw(PROP_WRITE_HANDLE, value)

    @property
    def mode(self):
        """Return the mode (auto / manual)."""
        return self._mode

    @property
    def mode_readable(self):
        """Return a readable representation of the mode.."""
        return self.decode_mode(self._mode)

    @property
    def min_temp(self):
        """Return the minimum temperature."""
        return 4.5

    @property
    def max_temp(self):
        """Return the maximum temperature."""
        return 30.5

    def pack_byte(self, byte):
        """Pack a byte."""
        return self._conn.pack_byte(byte)

    @staticmethod
    def decode_mode(mode):
        """Convert the numerical mode to a human-readable description."""
        ret = ""
        if mode & 1:
            ret = "manual"
        else:
            ret = "auto"

        if mode & 2:
            ret = ret + " holiday"
        if mode & 4:
            ret = ret + " boost"
        if mode & 8:
            ret = ret + " dst"
        if mode & 16:
            ret = ret + " window"

        return ret

    def update(self):
        """Update the data from the thermostat."""
        self._conn.write_request_raw(PROP_WRITE_HANDLE, self.pack_byte(PROP_INFO_QUERY))
